import io
import json
import contextlib
try:
    from types import TypeType as type
except ImportError:
    pass

import semver

from . import util


class PackageStub(object):
    keys = ['name', 'version', 'description', 'license', 'compatibility']

    def __init__(self, defaults=None):
        defaults = defaults or {}
        for key in self.keys:
            setattr(self, key, defaults.get(key))

    def is_valid(self, raise_exception=False):
        res = False
        if self.name and self.version:
            res = True

        if raise_exception and not res:
            raise Exception('invalid package')
        return res

    @property
    def ident(self):
        if self.is_valid(True):
            return util.archive_filename(self.name, self.version)

    def to_json(self, keys=None):
        return util.json_dump(self.to_dict(keys))

    def to_dict(self, keys=None):
        keys = keys or []
        if hasattr(self, 'is_valid'):
            self.is_valid()
        return dict([
            (k, getattr(self, k))
            for k in self.keys
            if not keys or k in keys])

    def has_file(self, *path_parts):
        raise NotImplementedError

    def file_path(self, *path_parts):
        raise NotImplementedError

    def dir_path(self, *path_parts):
        raise NotImplementedError

    @contextlib.contextmanager
    def open(self, path_parts, mode='r', encoding='utf8', default=IOError):
        if self.has_file(*path_parts):
            f = io.open(self.file_path(*path_parts),
                        mode=mode, encoding=encoding)
            yield f
            f.close()

        else:
            if isinstance(default, type) and issubclass(default, Exception):
                raise default(self.file_path(*path_parts))
            elif isinstance(default, Exception):
                raise default
            else:
                yield default

    def load_json(self, path_parts, mode='r', encoding='utf8', default=IOError):
        if self.has_file(*path_parts):
            with io.open(self.file_path(*path_parts),
                         mode=mode, encoding=encoding) as f:
                return json.load(f)

        else:
            if isinstance(default, type) and issubclass(default, Exception):
                raise default(self.file_path(*path_parts))
            elif isinstance(default, Exception):
                raise default
            else:
                return default

    def _error_on_different_name(self, other):
        if self.name != other.name:
            raise Exception('name mismatch: %s != %s' % (self.name, other.name))

    def __gt__(self, other):
        self._error_on_different_name(other)
        return semver.compare(self.version, other.version) > 0

    def __lt__(self, other):
        self._error_on_different_name(other)
        return semver.compare(self.version, other.version) < 0

    def __eq__(self, other):
        self._error_on_different_name(other)
        return semver.compare(self.version, other.version) == 0

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)



import os
import logging

from . import util
from . import default
from .package_stub import PackageStub


class NotIncludedException(Exception): pass


class Package(PackageStub):  # installed package
    def __init__(self, path):
        meta = util.json_load(os.path.join(path, default.META_FILENAME))
        super(Package, self).__init__(defaults=meta['package'])

        self.logger = logging.getLogger(__name__)
        self.meta = meta
        self.path = path

    @property
    def manifest(self):
        return self.meta['manifest']

    def has_file(self, *path_parts):
        return any(m for m in self.manifest if tuple(m['path']) == path_parts)

    def file_path(self, *path_parts):
        path = util.get_path(*path_parts)

        if not self.has_file(*path_parts):
            raise NotIncludedException('package does not include file: %s' % path)

        return os.path.join(self.path, path)

    def dir_path(self, *path_parts):
        # TODO check whether path is part of package
        path = util.get_path(*path_parts)
        return os.path.join(self.path, path)



import os
import subprocess
import shutil

import pytest

from .. import site


def pip_uninstall(package_name):
    subprocess.call(['pip', 'uninstall', '-yq', package_name])


def pip_install(package_name):
    subprocess.call(['pip', 'install', '-q', package_name])


@pytest.fixture
def package_name():
    return 'text_unidecode'


@pytest.fixture
def record_path():
    return site.get_record_path('setuptools')


def test_pip(package_name):
    pip_install(package_name)
    try:
        path = site.get_mod_path(package_name)
        assert os.path.exists(path)
    finally:
        pip_uninstall(package_name)
        assert not os.path.exists(path)


def test_get_record_path():
    assert site.get_record_path('setuptools')


def test_record_has_path(package_name, record_path):
    assert site.record_has_path(record_path, 'setuptools/__init__.py')


def test_record_has_path(record_path):
    assert site.record_has_path(record_path, 'setuptools/__init__.py')


def test_record_add_path(package_name):
    pip_install(package_name)
    try:
        path = site.get_mod_path(package_name)
        assert os.path.exists(path)

        record_path = site.get_record_path(package_name)
        assert record_path

        assert site.record_has_path(record_path, '%s/__init__.py' % package_name)
        assert not site.record_has_path(record_path, '%s/data' % package_name)
        site.record_add_path(record_path, '%s/data' % package_name)
        assert site.record_has_path(record_path, '%s/data' % package_name)
    finally:
        pip_uninstall(package_name)
        assert not os.path.exists(path)


def test_add_path(package_name):
    pip_install(package_name)
    try:
        path = site.get_mod_path(package_name)
        assert os.path.exists(path)

        record_path = site.get_record_path(package_name)
        assert record_path

        data_path = os.path.join(path, 'data')
        assert not os.path.exists(data_path)
        os.mkdir(data_path)

        assert not site.record_has_path(record_path, '%s/data' % package_name)
        site.add_path(package_name, 'data')
        assert site.record_has_path(record_path, '%s/data' % package_name)
    finally:
        pip_uninstall(package_name)
        assert not os.path.exists(path)


@pytest.mark.xfail
def test_add_outside_path(package_name):
    pip_install(package_name)
    try:
        path = site.get_mod_path(package_name)
        assert os.path.exists(path)

        record_path = site.get_record_path(package_name)
        assert record_path

        assert not site.record_has_path(record_path, 'data')
        site.add_path(record_path, '../data')
        assert not site.record_has_path(record_path, '../data')

        assert not os.path.exists(path)
    finally:
        pip_uninstall(package_name)
        assert not os.path.exists(path)
